from afmformats.formats import formats_by_mode

from . import fd


analysis_types = {
    "fd": {
        "suffixes": [f["suffix"] for f in formats_by_mode["force-distance"]],
        "gui": fd.UiForceDistance,
        }
    }

known_suffixes = []
for _item in analysis_types.values():
    known_suffixes += _item["suffixes"]



import hashlib
import io
import os
import pkg_resources
import time

import afmformats
import nanite
import nanite.fit as nfit
import numpy as np
from PyQt5 import uic, QtCore, QtGui, QtWidgets

from .. import colormap

from . import rating_base
from . import rating_iface
from . import export


# load QWidget from ui file
dlg_autosave_path = pkg_resources.resource_filename("pyjibe.fd",
                                                    "dlg_autosave_design.ui")
DlgAutosave = uic.loadUiType(dlg_autosave_path)[0]


class UiForceDistance(QtWidgets.QWidget):
    _instance_counter = 0

    def __init__(self, *args, **kwargs):
        """Base class for force-indentation analysis"""
        super(UiForceDistance, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename("pyjibe.fd", "main.ui")
        uic.loadUi(path_ui, self)

        UiForceDistance._instance_counter += 1
        title = "Force-Distance #{}".format(self._instance_counter)
        self.parent().setWindowTitle(title)

        self.data_set = nanite.IndentationGroup()

        # rating scheme
        self.rating_scheme_setup()

        # Signals
        # tabs
        self.tabs.currentChanged.connect(self.on_tab_changed)
        # fitting / parameters
        self.tab_edelta.sp_delta_num_samples.valueChanged.connect(
            self.on_params_init)
        self.btn_fitall.clicked.connect(self.on_fit_all)
        self.tab_fit.cb_delta_select.currentIndexChanged.connect(
            self.tab_edelta.cb_delta_select.setCurrentIndex)
        self.tab_edelta.cb_delta_select.currentIndexChanged.connect(
            self.tab_fit.cb_delta_select.setCurrentIndex)
        self.tab_fit.sp_range_1.valueChanged["double"].connect(
            self.tab_edelta.on_delta_change_spin)
        self.tab_fit.sp_range_2.valueChanged.connect(self.on_params_init)
        # rating
        self.btn_rating_filter.clicked.connect(self.on_rating_threshold)
        self.cb_rating_scheme.currentTextChanged.connect(
            self.on_cb_rating_scheme)
        self.btn_rater.clicked.connect(self.on_user_rate)
        # plotting parameters
        self.cb_mpl_rescale_plot_x.stateChanged.connect(
            self.on_mpl_curve_update)
        self.cb_mpl_rescale_plot_x_min.valueChanged.connect(
            self.on_mpl_curve_update)
        self.cb_mpl_rescale_plot_x_max.valueChanged.connect(
            self.on_mpl_curve_update)
        self.cb_mpl_rescale_plot_y.stateChanged.connect(
            self.on_mpl_curve_update)
        self.cb_mpl_rescale_plot_y_min.valueChanged.connect(
            self.on_mpl_curve_update)
        self.cb_mpl_rescale_plot_y_max.valueChanged.connect(
            self.on_mpl_curve_update)

        # Random string for identification of autosaved results
        self._autosave_randstr = hashlib.md5(time.ctime().encode("utf-8")
                                             ).hexdigest()[:5]
        # Initialize `override` parameter:
        # -1: not decided
        #  0: do not override existing files
        #  1: override existing files
        #  2: create additional file with date
        self._autosave_override = -1
        # Filenames that were created by this instance
        self._autosave_original_files = []

    @property
    def current_curve(self):
        idx = self.current_index
        fdist = self.data_set[idx]
        return fdist

    @property
    def current_index(self):
        item = self.list_curves.currentItem()
        idx = self.list_curves.indexOfTopLevelItem(item)
        return idx

    @property
    def selected_curves(self):
        """Return an IndentationGroup with all curves selected by the user"""
        curves = nanite.IndentationGroup()
        for ar in self.data_set:
            idx = self.data_set.index(ar)
            item = self.list_curves.topLevelItem(idx)
            if item.checkState(3) == 2:
                curves.append(ar)
        return curves

    def add_files(self, files):
        """ Populate self.data_set and display the first curve """
        # The `mult` parameter is used to chunk the progress bar,
        # because we cannot use floats with `QProgressDialog`, but
        # we want to be able to show progress for files that contain
        # multiple force curves.
        mult = 100
        bar = QtWidgets.QProgressDialog("Loading data files...",
                                        "Stop", 1, len(files)*mult)
        bar.setWindowTitle("Loading data files")
        bar.setMinimumDuration(1000)
        for ii, f in enumerate(files):
            label = "Loading file\n{}".format(os.path.basename(f))
            bar.setLabelText(label)

            def callback(partial):
                """Call back method for a progress dialog

                This method updates `bar` with the current file index
                `ii` and a partial file value `partial`.

                Parameters
                ----------
                partial: float in [0,1]
                    The progress for a single file
                """
                bar.setValue(int((ii+partial)*mult))
                QtCore.QCoreApplication.instance().processEvents()
                if bar.wasCanceled():
                    # Raise a custom `AbortProgress` error, such that
                    # we can exit the parent for-loop.
                    raise AbortProgress
            try:
                grp = nanite.IndentationGroup(f, callback=callback)
                callback(1)
            except afmformats.errors.FileFormatMetaDataError:
                # ignore e.g. JPK callibration curves
                callback(1)
                continue
            except AbortProgress:
                # The custom error `AbortProgress` was called, because
                # the user wants to stop loading the data.
                break
            self.data_set += grp
        bar.reset()
        bar.close()
        self.curve_list_setup()
        # Select first item
        it = self.list_curves.topLevelItem(0)
        self.list_curves.setCurrentItem(it)

    def autosave(self, fdist):
        """Performs autosaving for all files"""
        if (self.cb_autosave.checkState() == 2 and
            fdist.fit_properties and
                fdist.fit_properties["success"]):
            # Determine the directory of the current curve
            adir = os.path.dirname(fdist.path)
            model_key = fdist.fit_properties["model_key"]
            # Determine all other curves with the same path
            exp_curv = []
            for ii, ar in enumerate(self.data_set):
                it = self.list_curves.topLevelItem(ii)
                if (
                    # same directory
                    os.path.dirname(ar.path) == adir and
                    # fdist was fitted
                    ar.fit_properties and
                    # fit was successful
                    ar.fit_properties["success"] and
                    # fdist was fitted with same model
                    ar.fit_properties["model_key"] == model_key and
                    # user selected curve for export ("use")
                    it.checkState(3) == 2
                ):
                    exp_curv.append(ar)
            # The file to export
            fname = os.path.join(adir, "pyjibe_fit_results_leaf.tsv")

            # Only export if we have curves to export
            if exp_curv:
                if (os.path.exists(fname) and
                        fname not in self._autosave_original_files):
                    # File already exists
                    oride = self._autosave_override
                    if oride == -1:
                        # Ask user what to do
                        dlgwin = QtWidgets.QDialog(self)
                        dlgwin.setWindowModality(QtCore.Qt.ApplicationModal)
                        dlgui = DlgAutosave()
                        dlgui.setupUi(dlgwin)
                        if dlgwin.exec_():
                            if dlgui.btn_nothing.isChecked():
                                oride = 0
                            elif dlgui.btn_override.isChecked():
                                oride = 1
                            elif dlgui.btn_new.isChecked():
                                oride = 2
                            if dlgui.cb_remember.isChecked():
                                self._autosave_override = oride
                            if dlgui.cb_disableauto.isChecked():
                                self.cb_autosave.setChecked(0)
                    if oride == 0:
                        # Do not override
                        return
                    elif oride == 1:
                        # Override existing file
                        pass
                    elif oride == 2:
                        # Create new filename
                        now = time.strftime("%Y-%m-%d")
                        newbase = "pyjibe_fit_results_leaf_{}_{}.tsv".format(
                            now,
                            self._autosave_randstr
                        )
                        fname = os.path.join(adir, newbase)
                # Export data
                ratings = self.rate_data(exp_curv)
                export.save_tsv_approach_retract(filename=fname,
                                                 fdist_list=exp_curv,
                                                 ratings=ratings)
                self._autosave_original_files.append(fname)

    def curve_list_setup(self):
        """Add items to the tree widget"""
        header = self.list_curves.header()
        header.setStretchLastSection(False)
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        self.list_curves.setColumnWidth(1, 70)
        self.list_curves.setColumnWidth(2, 70)
        self.list_curves.setColumnWidth(3, 40)
        for ar in self.data_set:
            it = QtWidgets.QTreeWidgetItem(self.list_curves,
                                           ["..."+str(ar.path)[-62:],
                                            str(ar.enum),
                                            "{:.1f}".format(-1)])
            self.list_curves.addTopLevelItem(it)
            it.setCheckState(3, QtCore.Qt.Checked)
        # Connect signals:
        # Selection of curves
        self.list_curves.itemSelectionChanged.connect(self.on_curve_list)
        self.list_curves.model().dataChanged.connect(
            self.on_curve_list_item_changed)

    def curve_list_update(self, item=None):
        """Update the curve list display with all ratings"""
        if item is None:
            indices = np.arange(len(self.data_set))
        else:
            indices = [item]
        for ii in indices:
            ar = self.data_set[ii]
            rating = self.rate_data(ar)
            it = self.list_curves.topLevelItem(ii)
            it.setText(2, "{:.1f}".format(rating))
            cm = colormap.cm_rating
            color = np.array(cm(rating/10))*255
            it.setBackground(2, QtGui.QColor(*color))

    @staticmethod
    def get_export_choices():
        """Choices for file menu export

        Returns
        -------
        choices: list
            [[menu label 1, method name 1],
             ...
             ]
        """
        choices = [["fit results", "on_export_fit_results"],
                   ["E(δ) curves", "on_export_edelta"]
                   ]
        return choices

    def info_update(self, fdist):
        """Updates the info tab"""
        self.tab_info.update_info(fdist)

    def on_cb_rating_scheme(self):
        """Switch rating scheme or import a new one"""
        scheme_id = self.cb_rating_scheme.currentIndex()
        schemes = rating_base.get_rating_schemes()
        if len(schemes) == scheme_id:
            search_dir = ""
            exts_str = "Training set zip file (*.zip)"
            tsz, _e = QtWidgets.QFileDialog.getOpenFileName(
                self.parent(), "Import a training set",
                search_dir, exts_str)
            if tsz:
                idx = rating_base.import_training_set(tsz)
                self.rating_scheme_setup()
                self.cb_rating_scheme.setCurrentIndex(idx)
            else:
                self.cb_rating_scheme.setCurrentIndex(0)
        else:
            self.on_params_init()

    def on_curve_list(self):
        """Called when a new curve is selected"""
        fdist = self.current_curve
        idx = self.current_index
        # perform preprocessing
        self.tab_preprocess.fit_apply_preprocessing(fdist)
        # update user interface with initial parameters
        self.tab_fit.fit_update_parameters(fdist)
        # fit data
        self.tab_fit.fit_approach_retract(fdist)
        # set plot data (time consuming)
        self.widget_fdist.mpl_curve_update(fdist)
        # update info
        self.info_update(fdist)
        # Display new rating
        self.curve_list_update(item=idx)
        # Display map
        self.tab_qmap.mpl_qmap_update()
        # Display edelta
        self.tab_edelta.mpl_edelta_update()

    def on_curve_list_item_changed(self, item):
        """An item in the curve list was changed

        - update the qmap
        - autosave the results
        """
        # The data set index:
        idx = item.row()
        fdist = self.data_set[idx]
        if item.column() == 3:
            # The checkbox has been changed
            self.tab_qmap.mpl_qmap_update()
            self.autosave(fdist)

    def on_export_edelta(self):
        """Saves all edelta curves"""
        fname, _e = QtWidgets.QFileDialog.getSaveFileName(
            self.parent(),
            "Save E(δ) curves",
            "",
            "Tab Separated Values (*.tsv)"
        )
        if fname:
            if not fname.endswith(".tsv"):
                fname += ".tsv"
            # Make sure all curves have correct fit properties
            self.on_fit_all()
            # Proceed computing edelta curves with progress bar
            curves = self.selected_curves
            bar = QtWidgets.QProgressDialog("Computing  E(δ) curves...",
                                            "Stop", 1, len(curves))
            bar.setWindowTitle("Please wait...")
            bar.setMinimumDuration(1000)

            res = []
            for ii, ar in enumerate(curves):
                if bar.wasCanceled():
                    return
                # TODO:
                # - Use the callback method in `compute_emodulus_mindelta`
                #   to prevent "freezing" of the GUI?
                try:
                    e, d = ar.compute_emodulus_mindelta()
                except nfit.FitDataError:
                    pass
                else:
                    res += [d, e]
                QtCore.QCoreApplication.instance().processEvents()
                bar.setValue(ii+1)

            # export curves with numpy
            with io.open(fname, "w") as fd:
                header = ["# Indentation [m] and elastic modulus [Pa]\n",
                          "# are stored as alternating rows.\n",
                          ]
                fd.writelines(header)

            with io.open(fname, "ab") as fd:
                np.savetxt(fd, np.array(res))

    def on_export_fit_results(self):
        """Saves all fit results"""
        fname, _e = QtWidgets.QFileDialog.getSaveFileName(
            self.parent(),
            "Save fit results",
            "fit_results_{:03d}.tsv".format(
                self._instance_counter),
            "Tab Separated Values (*.tsv)"
        )
        if fname:
            if not fname.endswith(".tsv"):
                fname += ".tsv"
            self.on_fit_all()
            exp_curv = [fdist for fdist in self.selected_curves]
            ratings = self.rate_data(exp_curv)
            export.save_tsv_approach_retract(filename=fname,
                                             fdist_list=exp_curv,
                                             ratings=ratings)

    def on_fit_all(self):
        """Apply initial parameters to all curves and fit"""
        # We will fit all curves with the currently visible settings
        bar = QtWidgets.QProgressDialog("Fitting all curves...",
                                        "Stop", 1, len(self.data_set))
        bar.setWindowTitle("Loading data files")
        bar.setMinimumDuration(1000)
        for ii, fdist in enumerate(self.data_set):
            QtCore.QCoreApplication.instance().processEvents()
            if bar.wasCanceled():
                break
            self.tab_preprocess.fit_apply_preprocessing(fdist)
            self.tab_fit.fit_approach_retract(fdist, update_ui=False)
            bar.setValue(ii+1)
            self.curve_list_update(item=ii)
        # Display map
        self.tab_qmap.mpl_qmap_update()

    def on_model(self):
        """Called when the fitting model is changed"""
        # The difference to "on_params_init" is that we
        # have to `fit_update_parameters` in order to display
        # potential new parameter names of the new model.
        fdist = self.current_curve
        self.tab_preprocess.fit_apply_preprocessing(fdist)
        self.tab_fit.fit_update_parameters(fdist)
        self.tab_fit.fit_approach_retract(fdist)
        self.widget_fdist.mpl_curve_update(fdist)
        self.curve_list_update()
        self.tab_qmap.mpl_qmap_update()

    def on_mpl_curve_update(self):
        fdist = self.current_curve
        self.widget_fdist.mpl_curve_update(fdist)

    def on_params_init(self):
        """Called when the initial parameters are changed"""
        fdist = self.current_curve
        idx = self.current_index
        self.tab_preprocess.fit_apply_preprocessing(fdist)
        self.tab_fit.fit_approach_retract(fdist)
        self.widget_fdist.mpl_curve_update(fdist)
        self.curve_list_update(item=idx)
        self.tab_qmap.mpl_qmap_update()

    def on_rating_threshold(self):
        """(De)select curves according to threshold rating"""
        thresh = self.sp_rating_thresh.value()
        self.curve_list_update()
        for ii, ar in enumerate(self.data_set):
            rating = self.rate_data(ar)
            it = self.list_curves.topLevelItem(ii)
            if rating >= thresh:
                it.setCheckState(3, 2)
            else:
                it.setCheckState(3, 0)
        # TODO:
        # - make this more efficient. There is a lot written to disk here.
        for fdist in self.data_set:
            self.autosave(fdist)

    def on_tab_changed(self, index):
        """Called when the tab on the right hand is changed"""
        if hasattr(self, "user_tab_selected"):
            prevtab = self.user_tab_selected
        else:
            prevtab = self.tabs.currentWidget()

        curtab = self.tabs.currentWidget()
        if curtab == self.tab_fit and prevtab == self.tab_preprocess:
            self.on_params_init()
        elif curtab == self.tab_qmap:
            # Redraw the current map
            self.tab_qmap.mpl_qmap_update()
        elif curtab == self.tab_edelta:
            # Compute edelta plot
            self.tab_edelta.mpl_edelta_update()

        self.user_tab_selected = curtab

    def on_user_rate(self):
        """Start the curve rater"""
        cont = QtWidgets.QFileDialog.getSaveFileName(
            parent=None,
            caption="Please select a rating container",
            directory="",
            filter="Rating containers (*.h5)",
            options=QtWidgets.QFileDialog.DontConfirmOverwrite
                    | QtWidgets.QFileDialog.DontUseNativeDialog)

        path = cont[0]
        if path:
            rt = rating_iface.Rater(fdui=self, path=path)
            self.curve_rater = rt
            rt.show()

    def rate_data(self, data):
        """Apply rating to a force-distance curves (or a list of curves)"""
        scheme_id = self.cb_rating_scheme.currentIndex()
        return rating_base.rate_fdist(data, scheme_id)

    def rating_scheme_setup(self):
        self.cb_rating_scheme.clear()
        schemes = rating_base.get_rating_schemes()
        self.cb_rating_scheme.addItems(list(schemes.keys()))
        self.cb_rating_scheme.addItem("Add...")


class AbortProgress(BaseException):
    pass



import pkg_resources

import nanite
from PyQt5 import uic, QtWidgets

from .mpl_qmap import MPLQMap


class TabQMap(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(TabQMap, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename("pyjibe.fd",
                                                  "tab_qmap.ui")
        uic.loadUi(path_ui, self)

        # Setup the matplotlib interface for 2D map plotting
        self.mpl_qmap = MPLQMap()
        self.mpl_qmap.add_toolbar(self.qmap_mplwidget)
        self.qmap_mpllayout.addWidget(self.mpl_qmap.canvas)
        self.qmap_mpllayout.addWidget(self.mpl_qmap.toolbar)
        # set colormaps
        cmaps = ["viridis", "plasma", "afmhot", "seismic"]
        for cm in cmaps:
            self.qmpa_cmap_cb.addItem(cm)
        self.qmpa_cmap_cb.setCurrentIndex(0)
        self.qmap_data_cb.currentIndexChanged.connect(
            self.on_qmap_data_changed)
        self.qmpa_cmap_cb.currentIndexChanged.connect(
            self.on_qmap_cmap_changed)
        self.qmap_sp_range1.valueChanged.connect(self.on_qmap_min_max_changed)
        self.qmap_sp_range2.valueChanged.connect(self.on_qmap_min_max_changed)
        self.mpl_qmap.connect_curve_selection_event(self.on_qmap_selection)

    @property
    def current_curve(self):
        return self.fd.current_curve

    @property
    def fd(self):
        return self.parent().parent().parent().parent()

    def mpl_qmap_update(self):
        # Only update if we are on the right tab
        if self.fd.tabs.currentWidget() == self:
            fdist = self.current_curve
            # Get all selected curves with the same path
            fdist_map = self.fd.selected_curves.subgroup_with_path(fdist.path)
            self.update_qmap(fdist_map, fdist_map.index(fdist))

    def on_qmap_cmap_changed(self):
        """colormap selection changed"""
        self.mpl_qmap_update()

    def on_qmap_data_changed(self):
        """data column selection changed"""
        # set previous spin control values if existent
        self.qmap_sp_range1.blockSignals(True)
        self.qmap_sp_range2.blockSignals(True)
        if hasattr(self, "_cache_qmap_spin_ctl"):
            data = self.qmap_data_cb.currentIndex()
            if data in self._cache_qmap_spin_ctl:
                vmin, vmax = self._cache_qmap_spin_ctl[data]
            else:
                vmin = vmax = 0
            self.qmap_sp_range1.setValue(vmin)
            self.qmap_sp_range2.setValue(vmax)
        self.qmap_sp_range1.blockSignals(False)
        self.qmap_sp_range2.blockSignals(False)
        self.mpl_qmap_update()

    def on_qmap_min_max_changed(self):
        """min or max spin controls changed"""
        # store spin control values for data column
        vmin = self.qmap_sp_range1.value()
        vmax = self.qmap_sp_range2.value()
        data = self.qmap_data_cb.currentIndex()
        if not hasattr(self, "_cache_qmap_spin_ctl"):
            self._cache_qmap_spin_ctl = {}
        self._cache_qmap_spin_ctl[data] = (vmin, vmax)
        self.mpl_qmap_update()

    def on_qmap_selection(self, idx):
        """Show the curve indexed in the current qmap"""
        # Perform operations on ForceDistance
        fd = self.parent().parent().parent().parent()
        # Get the qmap name
        cc = fd.current_curve
        # idx is `enum` and curves are sorted
        curves = [ci for ci in fd.data_set if ci.path == cc.path]
        fdist = curves[idx]
        idcurve = fd.data_set.index(fdist)
        item = fd.list_curves.topLevelItem(idcurve)
        fd.list_curves.setCurrentItem(item)

    def update_qmap(self, fdist_map, index):
        # Build list of possible selections
        selist = nanite.qmap.available_features

        # Get plotting parameter and check if it makes sense
        feature = self.qmap_data_cb.currentText()
        if not feature or feature not in selist:
            # Use a default plotting map
            feature = "data min height"

        # Make sure that we have a valid property to plot
        assert feature in selist

        # Update dropdown menu with possible selections
        # disable signals while updating the combobox
        self.qmap_data_cb.blockSignals(True)
        # remove all items
        for _i in range(self.qmap_data_cb.count()):
            self.qmap_data_cb.removeItem(0)
        # add new items
        for item in selist:
            self.qmap_data_cb.addItem(item)
        self.qmap_data_cb.setCurrentIndex(selist.index(feature))
        self.qmap_data_cb.blockSignals(False)

        if len(fdist_map) > 1:
            # Get map data
            qmap = nanite.QMap(fdist_map)
            # update plot
            self.mpl_qmap.update(qmap=qmap,
                                 feature=feature,
                                 cmap=self.qmpa_cmap_cb.currentText(),
                                 vmin=self.qmap_sp_range1.value(),
                                 vmax=self.qmap_sp_range2.value())
            self.mpl_qmap.set_selection_by_index(index)
        else:
            self.mpl_qmap.reset()



import pathlib
import pkg_resources
import signal
import sys
import traceback
import webbrowser

from PyQt5 import uic, QtCore, QtWidgets

import afmformats
import appdirs
import h5py
import lmfit
import matplotlib
import nanite
import numpy
import scipy
import sklearn


from . import custom_widgets

from .. import registry
from ..settings import SettingsFile
from .._version import version as __version__


class PyJibeQMdiSubWindow(QtWidgets.QMdiSubWindow):
    def closeEvent(self, QCloseEvent):
        """Correctly de-register a data set before removing the subwindow"""
        mainwidget = self.mdiArea().parentWidget().parentWidget()
        mainwidget.rem_subwindow(self.windowTitle())
        super(PyJibeQMdiSubWindow, self).closeEvent(QCloseEvent)


class PyJibe(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs):
        super(PyJibe, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename("pyjibe.head", "main.ui")
        uic.loadUi(path_ui, self)

        self.setWindowTitle("PyJibe {}".format(__version__))
        # Disable native menubar (e.g. on Mac)
        self.menubar.setNativeMenuBar(False)
        # Connect menu entries
        # File menu
        self.action_open_bulk.triggered.connect(self.on_open_bulk)
        self.action_open_single.triggered.connect(self.on_open_single)
        self.action_open_multiple.triggered.connect(self.on_open_multiple)
        # Help menu
        self.actionDocumentation.triggered.connect(self.on_documentation)
        self.actionSoftware.triggered.connect(self.on_software)
        self.actionAbout.triggered.connect(self.on_about)
        # Add settings
        self.settings = SettingsFile()

        self.subwindows = []
        self.subwindow_data = []
        self.mdiArea.cascadeSubWindows()
        self.showMaximized()

    def add_subwindow(self, aclass, flist):
        """Add a subwindow, register data set and add to menu"""
        sub = PyJibeQMdiSubWindow()
        inst = aclass(sub)
        sub.setWidget(inst)
        inst.add_files(flist)
        self.mdiArea.addSubWindow(sub)
        sub.show()
        self.subwindows.append(sub)
        # Add export choices
        if hasattr(inst, "get_export_choices"):
            choices = inst.get_export_choices()
            menobj = self.menuExport.addMenu(sub.windowTitle())
            for choice in choices:
                action = menobj.addAction(choice[0])
                action.triggered.connect(getattr(inst, choice[1]))

    def rem_subwindow(self, title):
        """De-register a data set and remove from the menu"""
        for ii, sub in enumerate(self.subwindows):
            if sub.windowTitle() == title:
                self.subwindows.pop(ii)
                break

        for action in self.menuExport.actions():
            if action.text() == title:
                self.menuExport.removeAction(action)
                break

    def on_about(self):
        about_text = "PyJibe is a user interface for data analysis in " \
            + "atomic force microscopy with an emphasis on biological " \
            + "specimens, such as tissue sections or single cells.\n\n" \
            + "Author: Paul Müller\n" \
            + "Repository: https://github.com/AFM-analysis/PyJibe\n" \
            + "Documentation: https://pyjibe.readthedocs.io"
        QtWidgets.QMessageBox.about(self,
                                    "PyJibe {}".format(__version__),
                                    about_text)

    def on_documentation(self):
        webbrowser.open("https://pyjibe.readthedocs.io")

    def on_open_bulk(self, evt=None):
        dlg = custom_widgets.FileDialog(self)
        search_dir = self.settings.get_path("load data")
        dlg.setDirectory(search_dir)
        if dlg.exec_():
            files = dlg.selectedFilesRecursive()
            files.sort()
            if files:
                self.load_data(files=files, retry_open=self.on_open_bulk,
                               separate_analysis=False)
                self.settings.set_path(dlg.getDirectory(), name="load data")

    def on_open_multiple(self, evt=None):
        dlg = custom_widgets.FileDialog(self)
        search_dir = self.settings.get_path("load data")
        dlg.setDirectory(search_dir)

        if dlg.exec_():
            files = dlg.selectedFilesRecursive()
            files.sort()
            if files:
                self.load_data(files=files, retry_open=self.on_open_multiple,
                               separate_analysis=True)
                self.settings.set_path(dlg.getDirectory(), name="load data")

    def on_open_single(self, evt=None):
        exts = ["*"+e for e in registry.known_suffixes]
        exts_str = "Supported file types ({})".format(" ".join(exts))
        search_dir = self.settings.get_path("load data")

        n, _e = QtWidgets.QFileDialog.getOpenFileName(self, "Open single file",
                                                      search_dir, exts_str)
        if n:
            # user did not press cancel
            self.load_data(files=[n], retry_open=self.on_open_single)
            self.settings.set_path(pathlib.Path(n).parent, name="load data")

    def on_software(self):
        libs = [afmformats,
                appdirs,
                h5py,
                lmfit,
                matplotlib,
                nanite,
                numpy,
                sklearn,
                scipy,
                ]
        sw_text = "PyJibe {}\n\n".format(__version__)
        sw_text += "Python {}\n\n".format(sys.version)
        sw_text += "Modules:\n"
        for lib in libs:
            sw_text += "- {} {}\n".format(lib.__name__, lib.__version__)
        sw_text += "- PyQt5 {}\n".format(QtCore.QT_VERSION_STR)
        if hasattr(sys, 'frozen'):
            sw_text += "\nThis executable has been created using PyInstaller."
        QtWidgets.QMessageBox.information(self,
                                          "Software",
                                          sw_text)

    def load_data(self, files, retry_open=None, separate_analysis=False):
        # approach-retract data files
        supfiles = []
        for ff in files:
            path = pathlib.Path(ff)
            if path.suffix in registry.known_suffixes:
                supfiles.append(ff)

        # check if AFM files were found
        if not len(supfiles):
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setText("No AFM data files found!")
            msg.setInformativeText("")
            msg.setWindowTitle("No AFM data found!")
            msg.setStandardButtons(
                QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Retry)
            retval = msg.exec_()
            if (retry_open is not None and
                    retval == QtWidgets.QMessageBox.Retry):
                retry_open()
        else:
            if separate_analysis:
                # open each file in one analysis
                supfiles = [[ss] for ss in supfiles]
            else:
                supfiles = [supfiles]
            for flist in supfiles:
                aclass = registry.fd.UiForceDistance
                self.add_subwindow(aclass, flist)


def excepthook(etype, value, trace):
    """
    Handler for all unhandled exceptions.

    Parameters
    ----------
    etype: Exception
        exception type (`SyntaxError`, `ZeroDivisionError`, etc...)
    value: str
        exception error message
    trace:
        traceback header, if any (otherwise, it prints the
        standard Python header: ``Traceback (most recent call last)``
    """
    vinfo = "Unhandled exception in PyJibe version {}:\n".format(__version__)
    tmp = traceback.format_exception(etype, value, trace)
    exception = "".join([vinfo]+tmp)

    errorbox = QtWidgets.QMessageBox()
    errorbox.addButton(QtWidgets.QPushButton('Close'),
                       QtWidgets.QMessageBox.YesRole)
    errorbox.addButton(QtWidgets.QPushButton(
        'Copy text && Close'), QtWidgets.QMessageBox.NoRole)
    errorbox.setText(exception)
    ret = errorbox.exec_()
    if ret == 1:
        cb = QtWidgets.QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText(exception)


# Make Ctr+C close the app
signal.signal(signal.SIGINT, signal.SIG_DFL)
# Display exception hook in separate dialog instead of crashing
sys.excepthook = excepthook



import appdirs
import collections
import pathlib
import shutil
import zipfile

import nanite.indent as nindent
import nanite.rate.rater


#: Rating configuration directory
CFG_DIR = pathlib.Path(appdirs.user_config_dir(appname="PyJibe")) / "rating"
#: Path to main rating configuration file
CFG_PATH = CFG_DIR / "rating_schemes.txt"


def get_rating_schemes():
    """Return an ordered dict with available rating schemes"""
    schemes = collections.OrderedDict()
    ts = get_training_set_paths()
    # We currently stick to Extra Trees
    for key in ts:
        schemes["{} + Extra Trees".format(key)] = [ts[key], "Extra Trees"]
    schemes["Disabled"] = ["none", "none"]
    return schemes


def get_training_set_paths():
    """Return ordered dict with available training set names and paths"""
    ts = collections.OrderedDict()
    # training sets from nanite
    nanite_list = nanite.rate.rater.get_available_training_sets()
    for key in nanite_list:
        ts[key] = nanite.rate.IndentationRater.get_training_set_path(key)
    # user-imported training sets
    for pp in sorted(CFG_DIR.glob("ts_*")):
        ts[pp.name[3:]] = pp
    return ts


def import_training_set(ts_zip, override=False):
    """Open a training set zip file and import it to :const:`CFG_DIR`"""
    path = pathlib.Path(ts_zip)
    if not path.suffix == ".zip":
        raise ValueError("Training set file suffix must be '.zip', "
                         "got '{}'!".format(ts_zip.suffix))
    if not path.name.startswith("ts_"):
        raise ValueError("Training set file name must begin with 'ts_', "
                         "got '{}'!".format(ts_zip.name))
    pout = CFG_DIR / path.with_suffix("").name

    if pout.exists():
        if override:
            shutil.rmtree(pout)
        else:
            raise OSError("Training set already exists: {}".format(pout))

    pout.mkdir(exist_ok=True, parents=True)

    with zipfile.ZipFile(ts_zip) as zp:
        zp.extractall(pout)

    # return index in new training set collection
    ts = get_training_set_paths()
    for idx, key in enumerate(ts.keys()):
        if ts[key] == pout:
            break
    return idx


def rate_fdist(data, scheme_id):
    """Rate one or a list of force-distance curves

    Parameters
    ----------
    data: nanite.Indentation or a list of those
    """
    if isinstance(data, nindent.Indentation):
        data = [data]
        return_single = True
    else:
        return_single = False

    schemes = get_rating_schemes()
    scheme_key = list(schemes.keys())[scheme_id]
    training_set, regressor = schemes[scheme_key]
    rates = []
    for fdist in data:
        rt = fdist.rate_quality(regressor=regressor,
                                training_set=training_set)
        rates.append(rt)

    if return_single:
        return rates[0]
    else:
        return rates



import pkg_resources

import numpy as np
from PyQt5 import uic, QtWidgets

from .. import units
from .mpl_edelta import MPLEDelta


class TabEdelta(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(TabEdelta, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename("pyjibe.fd",
                                                  "tab_edelta.ui")
        uic.loadUi(path_ui, self)

        self.mpl_edelta_setup()

        # signals
        self.delta_spin.setValue(-np.inf)
        # Slider
        self.delta_slider.setMaximum(0)
        self.delta_spin.valueChanged["double"].connect(
            self.on_delta_change_spin)
        self.delta_slider.sliderReleasedFloat["double"].connect(
            self.on_delta_change_spin)
        self.delta_slider.valueChangedFloat["double"].connect(
            self.mpl_edelta.update_delta)
        self.delta_btn_guess.clicked.connect(self.on_delta_guess)
        self.sp_delta_num_samples.valueChanged.connect(self.mpl_edelta_update)

    @property
    def current_curve(self):
        return self.fd.current_curve

    @property
    def fd(self):
        return self.parent().parent().parent().parent()

    def mpl_edelta_setup(self):
        """Setup the matplotlib interface for E(delta) plotting"""
        self.mpl_edelta = MPLEDelta()
        self.mpl_edelta.add_toolbar(self.edelta_mplwidget)
        self.edelta_mpllayout.addWidget(self.mpl_edelta.canvas)
        self.edelta_mpllayout.addWidget(self.mpl_edelta.toolbar)

    def mpl_edelta_update(self):
        """Update the E(delta) plot"""
        if self.fd.tabs.currentWidget() == self:
            fdist = self.current_curve
            delta_opt = self.fd.tab_fit.sp_range_1.value()
            # Update slider range
            xaxis = self.fd.tab_fit.cb_xaxis.currentText()
            segment = self.fd.tab_fit.cb_segment.currentText().lower()
            segment_bool = segment == "retract"
            segid = fdist["segment"] == segment_bool
            xdata = fdist[xaxis]
            xscale = units.hrscale(xaxis)
            minx = np.min(xdata[segid])*xscale
            self.delta_slider.blockSignals(True)
            self.delta_slider.setMinimum(minx)
            self.delta_slider.setValue(delta_opt)
            self.delta_slider.blockSignals(False)
            # Update E(delta) plot
            self.fd.tab_fit.fit_approach_retract(fdist)
            self.mpl_edelta.update(fdist, delta_opt)

    def on_delta_guess(self):
        """Guess the optimal indentation depth for the current curve"""
        fdist = self.current_curve
        value = fdist.estimate_optimal_mindelta()
        value /= units.scales["µ"]
        self.delta_spin.setValue(value)

    def on_delta_change_spin(self, value):
        """Indentation depth spin control value changed"""
        # Update all controls
        self.fd.tab_fit.sp_range_1.blockSignals(True)
        self.delta_spin.blockSignals(True)
        self.delta_slider.blockSignals(True)

        self.delta_spin.setValue(value)
        self.delta_slider.setValue(value)
        self.fd.tab_fit.sp_range_1.setValue(value)
        self.mpl_edelta.update_delta(value)
        self.fd.on_params_init()

        self.fd.tab_fit.sp_range_1.blockSignals(False)
        self.delta_spin.blockSignals(False)
        self.delta_slider.blockSignals(False)

        # Update stored indentation depth
        if self.fd.tab_fit.cb_delta_select.currentIndex() == 1:
            # Set indentation depth individually
            fdist = self.current_curve
            self.fd.tab_fit._indentation_depth_individual[fdist] = value



import numbers
import pkg_resources

from afmformats import meta
from PyQt5 import uic, QtWidgets


class TabInfo(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(TabInfo, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename("pyjibe.fd",
                                                  "tab_info.ui")
        uic.loadUi(path_ui, self)

    def update_info(self, fdist):
        text = []
        keys = list(fdist.metadata.keys())

        text.append("<b>Dataset</b>")
        data_keys = ["path", "enum"]
        for kk in data_keys:
            text.append(get_string_rep(kk, fdist.metadata))
        for kk in keys:
            if kk in meta.DEF_DATA and kk not in data_keys:
                text.append(get_string_rep(kk, fdist.metadata))

        if set(keys) & set(meta.DEF_EXPERIMENT.keys()):
            text.append("")
            text.append("<b>Experiment:</b>")
            text_meta = []
            for kk in keys:
                if kk in meta.DEF_EXPERIMENT:
                    text_meta.append(get_string_rep(kk, fdist.metadata))
            text += sorted(text_meta)

        if set(keys) & set(meta.DEF_QMAP.keys()):
            text.append("")
            text.append("<b>QMap:</b>")
            text_meta = []
            for kk in keys:
                if kk in meta.DEF_QMAP:
                    text_meta.append(get_string_rep(kk, fdist.metadata))
            text += sorted(text_meta)

        if set(keys) & set(meta.DEF_ANALYSIS.keys()):
            text.append("")
            text.append("<b>Analysis:</b>")
            text_meta = []
            for kk in keys:
                if kk in meta.DEF_ANALYSIS:
                    text_meta.append(get_string_rep(kk, fdist.metadata))
            text += sorted(text_meta)

        textstring = "<br>".join(text)

        self.info_text.setText(textstring)


def get_string_rep(key, metadata):
    """Return a nice string representation for a key in metadata"""
    value = metadata[key]
    desc, unit, validator = meta.DEF_ALL[key]

    value = validator(value)

    if unit == "m":
        unit = "µm"
        value *= 1e6

    if isinstance(value, numbers.Integral):
        pass
    elif isinstance(value, numbers.Real):
        if abs(value) > 1:
            value = "{:.2f}".format(value)
        elif abs(value) > 1e-2:
            value = "{:.5f}".format(value)
        else:
            value = "{:.3e}".format(value)

    rep = "{}: {}".format(desc, value)
    if unit:
        rep += " {}".format(unit)
    return rep



import numpy as np

from PyQt5 import QtWidgets, QtCore


class InfDoubleSlider(QtWidgets.QSlider):
    valueChangedFloat = QtCore.pyqtSignal("double")
    sliderReleasedFloat = QtCore.pyqtSignal("double")

    def __init__(self, *args, **kwargs):
        super(InfDoubleSlider, self).__init__(*args, **kwargs)
        self._max_int = 100000

        super(InfDoubleSlider, self).setMinimum(0)
        super(InfDoubleSlider, self).setMaximum(self._max_int)

        self._min_value = -10
        self._max_value = 10

        self.valueChanged.connect(self.on_value_changed)
        self.sliderReleased.connect(self.on_slider_released)

    @property
    def _value_range(self):
        return self._max_value - self._min_value

    def value(self):
        pos = super(InfDoubleSlider, self).value()
        if pos == 0:
            value = -np.inf
        elif pos == self._max_int:
            value = np.inf
        else:
            # remove 2 because of inf borders
            value = (pos-1) / (self._max_int) * \
                self._value_range + self._min_value
        return value

    def on_slider_released(self):
        """Emits a double signal when `valueChanged["int"]` is emitted"""
        self.sliderReleasedFloat["double"].emit(self.value())

    def on_value_changed(self):
        """Emits a double signal when `valueChanged["int"]` is emitted"""
        self.valueChangedFloat["double"].emit(self.value())

    def setValue(self, value):
        if value == -np.inf:
            pos = 0
        elif value == np.inf:
            pos = self._max_int
        else:
            pos = int((value - self._min_value) /
                      self._value_range * (self._max_int))

        if pos < 0:
            pos = 0
        if pos > self._max_int:
            pos = self._max_int
        super(InfDoubleSlider, self).setValue(pos)

    def setMinimum(self, value):
        if value > self._max_value:
            raise ValueError("Minimum limit cannot be higher than maximum")
        orig_value = self.value()
        self._min_value = value
        self.setValue(orig_value)

    def setMaximum(self, value):
        if value < self._min_value:
            raise ValueError("Minimum limit cannot be higher than maximum")
        orig_value = self.value()
        self._max_value = value
        self.setValue(orig_value)

    def minimum(self):
        return self._min_value

    def maximum(self):
        return self._max_value



"""Widget containing force-distance plot"""
from PyQt5 import QtWidgets

from .mpl_indent import MPLIndentation


class WidgetFDist(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        """Display force-indentation graph with navigation"""
        super(WidgetFDist, self).__init__(*args, **kwargs)

        self.mplvl = QtWidgets.QVBoxLayout(self)
        # Setup the matplotlib interface for approach retract plotting
        self.mpl_curve = MPLIndentation()
        self.mpl_curve.add_toolbar(self)
        self.mplvl.addWidget(self.mpl_curve.canvas)
        self.mplvl.addWidget(self.mpl_curve.toolbar)

    @property
    def fd(self):
        return self.parent().parent()

    def mpl_curve_update(self, fdist):
        """Update the force-indentation curve"""
        autoscale_x = self.fd.cb_mpl_rescale_plot_x.checkState() == 2
        autoscale_y = self.fd.cb_mpl_rescale_plot_y.checkState() == 2
        if autoscale_x:
            rescale_x = None
        else:
            rescale_x = (self.fd.cb_mpl_rescale_plot_x_min.value(),
                         self.fd.cb_mpl_rescale_plot_x_max.value())

        if autoscale_y:
            rescale_y = None
        else:
            rescale_y = (self.fd.cb_mpl_rescale_plot_y_min.value(),
                         self.fd.cb_mpl_rescale_plot_y_max.value())

        self.mpl_curve.update(fdist,
                              rescale_x=rescale_x,
                              rescale_y=rescale_y)



import inspect
import pkg_resources

import nanite.preproc as npreproc
from PyQt5 import uic, QtWidgets


class TabPreprocess(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(TabPreprocess, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename("pyjibe.fd",
                                                  "tab_preprocess.ui")
        uic.loadUi(path_ui, self)

        # Setup everything necessary for the preprocessing tab:
        # Get list of preprocessing methods
        mem = inspect.getmembers(npreproc.IndentationPreprocessor)
        premem = []
        for m in mem:
            if (m[1].__doc__ is not None and
                not m[0].startswith("_") and
                    not m[0] == "apply"):
                premem.append(m[0])

        premem.sort()

        for p in premem:
            item = QtWidgets.QListWidgetItem()
            item.setText(p)
            self.list_preproc_available.addItem(item)

        self.list_preproc_available.currentItemChanged.connect(
            lambda: self.update_displayed_docstring("available"))
        self.list_preproc_applied.currentItemChanged.connect(
            lambda: self.update_displayed_docstring("applied"))

        # Add recommended item (see `self.preproc_set_preset`)
        self.cb_preproc_presel.addItem("Recommended")
        self.cb_preproc_presel.activated.connect(self.on_preset_changed)
        self.cb_preproc_presel.currentIndexChanged.connect(
            self.on_preset_changed)
        # Apply recommended defaults
        self.cb_preproc_presel.setCurrentIndex(1)

    def fit_apply_preprocessing(self, fdist):
        """Apply the preprocessing steps if required"""
        # Note: Preprocessing is cached once in `fdist`.
        # Thus calling this method a second time without any
        # change in the GUI is free.
        num = self.list_preproc_applied.count()
        preprocessing = []
        for ii in range(num):
            item = self.list_preproc_applied.item(ii)
            preprocessing.append(item.text())
        # Perform preprocessing
        fdist.apply_preprocessing(preprocessing)

    def on_preset_changed(self):
        """Update preselection"""
        text = self.cb_preproc_presel.currentText()
        self.list_preproc_applied.clear()
        if text == "None":
            pass
        elif text == "Recommended":
            recommended_methods = ["compute_tip_position",
                                   "correct_force_offset",
                                   "correct_tip_offset",
                                   "correct_split_approach_retract"]
            for m in recommended_methods:
                item = QtWidgets.QListWidgetItem()
                item.setText(m)
                self.list_preproc_applied.addItem(item)

    def update_displayed_docstring(self, source):
        """Update the description box with the method doc string

        Writes text to `self.text_preprocessing` according to the
        method selected in `self.list_preproc_available` or
        `self.list_preproc_applied`.
        """
        if source == "available":
            thelist = self.list_preproc_available
        elif source == "applied":
            thelist = self.list_preproc_applied
        item = thelist.currentItem()
        if item is None:
            return
        method_name = item.text()
        method = getattr(npreproc.IndentationPreprocessor, method_name)
        doc = method.__doc__.replace("    ", "").strip()
        self.text_preprocessing.setPlainText(doc)
